// (c) 2009 Uri Wilensky. See README.txt for terms of use.

package org.nlogo.lab

import java.util.concurrent.{Callable,ExecutionException,Executors,TimeUnit}
import org.nlogo.api.{CompilerException,Dump,LogoException,ValueConstraint,
                      WorldDimensions,WorldDimensionException}
import org.nlogo.nvm.{LabInterface,Workspace}
import LabInterface.ProgressListener

class Worker(val protocol: Protocol)
  extends LabInterface.Worker
{
  val listeners = new collection.mutable.ListBuffer[ProgressListener]
  def addListener(listener: ProgressListener) {
    listeners += listener
  }
  def addSpreadsheetWriter(modelFileName: String, initialDims: WorldDimensions, w: java.io.PrintWriter) {
    addListener(new SpreadsheetExporter(modelFileName, initialDims, protocol, w))
  }
  def addTableWriter(modelFileName: String, initialDims: WorldDimensions, w: java.io.PrintWriter) {
    addListener(new TableExporter(modelFileName, initialDims, protocol, w))
  }
  var runners: Seq[Runner] = null
  // we only want to compile stuff once per workspace, so use this
  // (should use a Scala collection not a Java one, but oh well, too lazy today - ST 8/13/09)
  val proceduresMap = new java.util.WeakHashMap[Workspace, Procedures]
  def run(initialWorkspace: Workspace, fn: ()=>Workspace, threads: Int) {
    try {
      val executor = Executors.newFixedThreadPool(threads)
      listeners.foreach(_.experimentStarted())
      runners =
        (for((settings, runNumber) <- protocol.elements zip Stream.from(1).elements)
         yield new Runner(runNumber, settings, fn)).collect
      val futures =
        org.nlogo.util.JCL.iterableFromJava(
          executor.invokeAll(
            org.nlogo.util.JCL.seqToJavaList[Callable[Unit]](
              runners)))
      executor.shutdown()
      executor.awaitTermination(java.lang.Integer.MAX_VALUE, TimeUnit.SECONDS)
      listeners.foreach(_.experimentCompleted())
      // this will cause the first ExecutionException we got to be thrown - ST 3/10/09
      futures.foreach(_.get)
    }
    catch { case _: InterruptedException => listeners.foreach(_.experimentAborted()) }
    finally { runners = null }
  }
  def compile(w: Workspace) { new Procedures(w) }  // for TestCompileAll. result discarded
  def abort() { if(runners != null) runners.foreach(_.aborted = true) }
  class Procedures(workspace: Workspace) {
    val setupProcedure = workspace.compileCommands(protocol.setupCommands)
    val goProcedure = workspace.compileCommands(protocol.goCommands
                                                + "\n" // protect against comments
                                                + "__experimentstepend")
    val finalProcedure = workspace.compileCommands(protocol.finalCommands)
    val exitProcedure =
      if(protocol.exitCondition.trim == "") None
      else Some(workspace.compileReporter(protocol.exitCondition))
    val metricProcedures = protocol.metrics.map(workspace.compileReporter(_))
  }
  class Runner(runNumber: Int, settings: List[Pair[String,Any]], fn: ()=>Workspace)
    extends Callable[Unit]
  {
    class FailedException(message: String) extends LogoException(message)
    @volatile var aborted = false
    // each Runner is on its own thread, but all the Runners share a ProgressListener,
    // so we need to synchronize
    def eachListener(fn: (ProgressListener)=>Unit) {
      listeners.synchronized { listeners.foreach(fn) }
    }
    def call() {
      val workspace = fn.apply
      try { callHelper(workspace) }
      catch { case t: Throwable =>
        if(!aborted) eachListener(_.runtimeError(workspace, runNumber, t)) }
    }
    def callHelper(workspace: Workspace) {
      val procedures =
        if(proceduresMap.containsKey(workspace))
          proceduresMap.get(workspace)
        else {
          val newProcedures = new Procedures(workspace)
          proceduresMap.put(workspace, newProcedures)
          newProcedures
        }
      import procedures._
      def setVariables(settings: List[Pair[String,Any]]) {
        val world = workspace.world
        var d = world.getDimensions
        for((name, value) <- settings) {
          if(world.isDimensionVariable(name)) {
            val v = value.asInstanceOf[java.lang.Double].intValue
            try { d = world.setDimensionVariable(name, v, d) }
            catch {
              case e: WorldDimensionException =>
                throw new FailedException("You cannot set " + name + " to " + v)
            }
          }
          else if(name.equalsIgnoreCase("RANDOM-SEED"))
            workspace.world.mainRNG.setSeed(value.asInstanceOf[java.lang.Double].longValue)
        }
        if(!world.equalDimensions(d)) workspace.setDimensions(d)
        for((name, value) <- settings)
          if(!world.isDimensionVariable(name) && !name.equalsIgnoreCase("RANDOM-SEED"))
            workspace.world.synchronized {
              if(workspace.world.observerOwnsIndexOf(name.toUpperCase) == -1)
                throw new FailedException(
                  "Global variable does not exist:\n" + name)
              workspace.world.setObserverVariableByName(name, value)
            }
      }
      def exitConditionTrue =
        exitProcedure match {
          case None => false
          case Some(procedure) =>
            workspace.runCompiledReporter(procedure, workspace.world.mainRNG.clone) match {
              case t: Throwable =>
                throw t
              case b: java.lang.Boolean =>
                b.booleanValue
              case null => 
                throw new FailedException(
                  "Stopping condition failed to report a result:\n" +
                  protocol.exitCondition)
              case result: AnyRef =>
                throw new FailedException(
                  "Stopping condition should report true or false, but instead reported the " +
                  Dump.typeName(result) + " " + Dump.logoObject(result))
            }
        }
      def takeMeasurements(): List[AnyRef] =
        metricProcedures.map{proc =>
          val result = workspace.runCompiledReporter(proc, workspace.world.mainRNG.clone)
          if(result == null)
            throw new FailedException(
              "Reporter for measuring runs failed to report a result:\n" + result)
          result }
      setVariables(settings)
      eachListener(_.runStarted(workspace, runNumber, settings))
      workspace.runCompiledCommands(setupProcedure)
      if(protocol.runMetricsEveryStep) {
        val m = takeMeasurements()
        eachListener(_.measurementsTaken(workspace, runNumber, 0, m))
      }
      var steps = 0
      while((protocol.timeLimit == 0 || steps < protocol.timeLimit) &&
            !exitConditionTrue && !workspace.runCompiledCommands(goProcedure))
      {
        steps += 1
        eachListener(_.stepCompleted(workspace, steps))
        if(protocol.runMetricsEveryStep && !listeners.isEmpty) {
          val m = takeMeasurements()
          eachListener(_.measurementsTaken(workspace, runNumber, steps, m))
        }
        workspace.updateDisplay(false)
        if(aborted) return
      }
      if(!protocol.runMetricsEveryStep) {
        val m = takeMeasurements()
        eachListener(_.measurementsTaken(workspace, runNumber, steps, m))
      }
      workspace.runCompiledCommands(finalProcedure)
      eachListener(_.runCompleted(workspace, runNumber, steps))
      if(workspace.lastLogoException != null) {
        val ex = workspace.lastLogoException
        workspace.clearLastLogoException()
        throw ex
      }
    }
  }
}
