//
// Copyright (c) 2000 Daniel Edelson. All rights reserved.
//

package org.myworldgis.io.shapefile;

import java.io.EOFException;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.ArrayList;
import java.util.List;
import org.nlogo.api.Syntax;


/**
 * Class representing a dBASE file.
 */
public final strictfp class DBaseFileReader implements DBaseConstants {
        
    //--------------------------------------------------------------------------
    // Instance variables
    //--------------------------------------------------------------------------
    
    /** */
    private RandomAccessFile _file;
    
    /** */
    private DBaseBuffer _buffer;
    
    /** */
    private DBaseFieldDescriptor[] _fieldDescriptors;
    
    /** */
    private int _headerSize;
    
    /** */
    private int _recordCount;
    
    /** */
    private int _recordLength;
    
    //--------------------------------------------------------------------------
    // Constructors
    //--------------------------------------------------------------------------
    
    /** */
    public DBaseFileReader (RandomAccessFile file) throws IOException {
        _file = file;
        _buffer = new DBaseBuffer();
        
        // read the header
        _file.seek(DBF_HEADER_SIZE_OFFSET);
        _buffer.read(_file, 0, 2);
        _headerSize = _buffer.getShort(0);
        _file.seek(0);
        int headerBytesRead = _buffer.read(_file, 0, _headerSize);
        if (headerBytesRead < _headerSize) {
            throw(new EOFException());
        }
        byte fileCode = _buffer.getByte(DBF_FILE_CODE_OFFSET);
        
        if (fileCode != DBF_FILE_CODE) {
            throw new IOException("invalid file code "+fileCode+", probably not a dBase file");
        }
        _recordCount = _buffer.getInt(DBF_RECORD_COUNT_OFFSET);
        _recordLength = _buffer.getShort(DBF_RECORD_SIZE_OFFSET);
        
        // read the field descriptors
        _file.seek(DBF_HEADER_SIZE);
        List<DBaseFieldDescriptor> fieldList = new ArrayList<DBaseFieldDescriptor>();
        while (true) {
            int fieldBytesRead = _buffer.read(_file, 0, DBF_FIELD_DESCRIPTOR_SIZE);
            if (_buffer.getByte(0) == DBF_FIELD_TERMINATOR) {
                break;
            } else if (fieldBytesRead < DBF_FIELD_DESCRIPTOR_SIZE) {
                throw new EOFException("unexpected end of file while reading field descriptors");
            }
            fieldList.add(_buffer.getFieldDescriptor(0));
        }
        _fieldDescriptors = fieldList.toArray(new DBaseFieldDescriptor[fieldList.size()]);
        
        _file.seek(_headerSize);
        _buffer.ensureCapacity(_recordLength);
    }
    
    //--------------------------------------------------------------------------
    // Instance methods
    //--------------------------------------------------------------------------
    
    /** */
    public int getFieldCount () {
        return _fieldDescriptors.length;
    }
    
    /** */
    public String getFieldName (int index) {
        return _fieldDescriptors[index].getName();
    }
    
    /** */
    public DBaseFieldDescriptor getField (int index) {
        return _fieldDescriptors[index];
    }
    
    /** */
    public int getFieldDataType (int index) {
        switch (_fieldDescriptors[index].getType()) {
            case DBaseConstants.FIELD_TYPE_CHARACTER:
            case DBaseConstants.FIELD_TYPE_DATE:
                return Syntax.TYPE_STRING;
            case DBaseConstants.FIELD_TYPE_NUMBER:
            case DBaseConstants.FIELD_TYPE_FLOAT:
                return Syntax.TYPE_NUMBER;
            case DBaseConstants.FIELD_TYPE_LOGICAL:
                return Syntax.TYPE_BOOLEAN;
            default:    
                return Syntax.TYPE_NOBODY;
        }
    }
    
    /** */
    public int getRecordCount () {
        return _recordCount;
    }
    
    /** */
    public Object[] getRecord (int index) throws IOException {
        _file.seek(_headerSize + (_recordLength * index));
        _buffer.read(_file, 0, _recordLength);
        return _buffer.getRecord(0, _fieldDescriptors);
    }
    
    /** */
    public void close () throws IOException {
        _file.close();
    }
}
