//
// Copyright (c) 2007 Eric Russell. All rights reserved.
//

package org.myworldgis.netlogo;

import com.vividsolutions.jts.geom.Envelope;
import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.util.GeometryTransformer;
import java.awt.image.BandedSampleModel;
import java.awt.image.DataBuffer;
import java.awt.image.Raster;
import java.awt.image.WritableRaster;
import java.io.*;
import java.text.ParseException;
import javax.measure.converter.UnitConverter;
import org.myworldgis.io.asciigrid.AsciiGridFileReader;
import org.myworldgis.io.shapefile.DBaseFileReader;
import org.myworldgis.io.shapefile.ESRIShapefileReader;
import org.myworldgis.projection.Projection;
import org.myworldgis.projection.ProjectionFormat;
import org.myworldgis.util.StringUtils;
import org.nlogo.api.*;


/**
 * 
 */
public final strictfp class LoadDataset extends GISExtension.Reporter {
    
    //--------------------------------------------------------------------------
    // Class methods
    //--------------------------------------------------------------------------
    
    /** */
    private static Dataset loadShapefile (String shpFile,
                                          Projection srcProj,
                                          Projection dstProj) throws IOException {
        GeometryTransformer inverse = null;
        GeometryTransformer forward = null;
        boolean reproject = false;
        if ((srcProj != null) && 
            (dstProj != null) &&
            (!srcProj.equals(dstProj))) {
            inverse = srcProj.getInverseTransformer();
            forward = dstProj.getForwardTransformer();
            reproject = true;
        }
        String dbfFile = StringUtils.changeFileExtension(shpFile, "dbf");
        ESRIShapefileReader shp = null;
        DBaseFileReader dbf = null;
        try {
            shp = new ESRIShapefileReader(new RandomAccessFile(shpFile, "r"), 
                                          UnitConverter.IDENTITY,
                                          GISExtension.getState().factory());
            dbf = new DBaseFileReader(new RandomAccessFile(dbfFile, "r"));
            VectorDataset.ShapeType shapeType = null;
            switch (shp.getShapeType()) {
                case ESRIShapefileReader.SHAPE_TYPE_POINT:
                case ESRIShapefileReader.SHAPE_TYPE_MULTIPOINT:
                    shapeType = VectorDataset.ShapeType.POINT;
                    break;
                case ESRIShapefileReader.SHAPE_TYPE_POLYLINE:
                    shapeType = VectorDataset.ShapeType.LINE;
                    break;
                case ESRIShapefileReader.SHAPE_TYPE_POLYGON:
                    shapeType = VectorDataset.ShapeType.POLYGON;
                    break;
                default:
                    throw new IOException("unsupported shape type " + shp.getShapeType());
            }
            String[] propertyNames = new String[dbf.getFieldCount()];
            VectorDataset.PropertyType[] propertyTypes = new VectorDataset.PropertyType[propertyNames.length];
            for (int i = 0; i < dbf.getFieldCount(); i += 1) {
                propertyNames[i] = dbf.getFieldName(i);
                if (dbf.getFieldDataType(i) == Syntax.TYPE_NUMBER) {
                    propertyTypes[i] = VectorDataset.PropertyType.NUMBER;
                } else {
                    propertyTypes[i] = VectorDataset.PropertyType.STRING;
                }
            }
            VectorDataset result = new VectorDataset(shapeType, propertyNames, propertyTypes);
            int recordIndex = 0;
            while (true) {
                Geometry shape = shp.getNextShape();
                if (shape == null) {
                    break;
                } else if (reproject) {
                    shape = forward.transform(inverse.transform(shape));
                }
                result.add(shape, dbf.getRecord(recordIndex++));
            }
            
            return result;
        } finally {
            if (shp != null) {
                try { shp.close(); } catch (IOException e) { 
                    // whose bright idea was it to allow close() 
                    // to throw an exception, anyway?
                }
            }
            if (dbf != null) {
                try { dbf.close(); } catch (IOException e) { }
            }
        }
    }
    
    /** */
    private static RasterDataset loadAsciiGrid (Reader in, 
                                               Projection srcProj,
                                               Projection dstProj) throws IOException {
        AsciiGridFileReader asc = null;
        try {
            asc = new AsciiGridFileReader(in);
            GridDimensions dimensions = new GridDimensions(asc.getSize(), asc.getEnvelope());
            DataBuffer data = asc.getData();
            BandedSampleModel sampleModel = new BandedSampleModel(data.getDataType(), 
                                                                  dimensions.getGridWidth(), 
                                                                  dimensions.getGridHeight(), 
                                                                  1);
            WritableRaster raster = Raster.createWritableRaster(sampleModel, data, null);
            if ((srcProj != null) && 
                (dstProj != null) &&
                (!srcProj.equals(dstProj))) {
                return new RasterDataset(raster, dimensions, srcProj, dstProj);
            } else {
                return new RasterDataset(dimensions, raster);
            }
        } finally {
            if (asc != null) {
                try { asc.close(); } catch (IOException e) { }
            }
        }
    }
    
    //--------------------------------------------------------------------------
    // GISExtension.Reporter implementation
    //--------------------------------------------------------------------------
    
    /** */
    public String getAgentClassString() {
        return "OTPL";
    }

    /** */
    public Syntax getSyntax() {
        return Syntax.reporterSyntax(new int[] { Syntax.TYPE_STRING },
                                     Syntax.TYPE_WILDCARD);
    }
    
    /** */
    public Object reportInternal (Argument args[], Context context) 
            throws ExtensionException, IOException, LogoException, ParseException {
        String fileName = args[0].getString();
        String dataFile = context.attachCurrentDirectory(fileName);
        java.io.File prjFile = new java.io.File(StringUtils.changeFileExtension(dataFile, "prj"));
        Projection netLogoProjection = GISExtension.getState().getProjection();
        Projection datasetProjection = null;
        if (prjFile.exists()) {
            Reader in = new BufferedReader(new FileReader(prjFile));
            try {
                datasetProjection = ProjectionFormat.getInstance().parseProjection(in);
            } finally {
                in.close();
            }
        }
        String extension = StringUtils.getFileExtension(dataFile);
        Dataset result = null;
        if (extension.equalsIgnoreCase(ESRIShapefileReader.SHAPEFILE_EXTENSION)) {
            result = loadShapefile(dataFile, datasetProjection, netLogoProjection);
        } else if (extension.equalsIgnoreCase(AsciiGridFileReader.ASCII_GRID_FILE_EXTENSION_1) ||
        		   extension.equalsIgnoreCase(AsciiGridFileReader.ASCII_GRID_FILE_EXTENSION_2)) {
            result = loadAsciiGrid(new FileReader(dataFile), datasetProjection, netLogoProjection);
        } else {
            throw new ExtensionException("unsupported file type "+extension);
        }
        // If the transformation hasn't been set yet, set it to map this
        // dataset's envelope to the NetLogo world using constant scale.
        if (!GISExtension.getState().isTransformationSet()) {
            World w = context.getAgent().world();
            GISExtension.getState().setTransformation(new CoordinateTransformation
													  (result.getEnvelope(),
													   new Envelope(w.minPxcor(), w.maxPxcor(), w.minPycor(), w.maxPycor()),
													   true));
        }
        return result;
    }
}
